import dateutil
import json

from cvss import parser as cvss_parser
from dojo.models import Finding


class GithubVulnerabilityParser(object):
    def get_scan_types(self):
        return ["Github Vulnerability Scan"]

    def get_label_for_scan_types(self, scan_type):
        return scan_type  # no custom label for now

    def get_description_for_scan_types(self, scan_type):
        return "Import vulnerabilities from Github API (GraphQL Query)"

    def get_findings(self, filename, test):
        data = json.load(filename)
        if "data" not in data:
            raise ValueError("Invalid report file, no 'data' node found")

        vulnerabilityAlerts = self._search_vulnerability_alerts(data["data"])
        if not vulnerabilityAlerts:
            raise ValueError(
                "Invalid report, no 'vulnerabilityAlerts' node found"
            )

        repository_url = None
        if "repository" in data["data"]:
            if "nameWithOwner" in data["data"]["repository"]:
                repository_url = "https://github.com/{}".format(
                    data["data"]["repository"]["nameWithOwner"]
                )
            if "url" in data["data"]["repository"]:
                repository_url = data["data"]["repository"]["url"]

        dupes = dict()
        for alert in vulnerabilityAlerts["nodes"]:
            description = alert["securityVulnerability"]["advisory"][
                "description"
            ]

            if "number" in alert and repository_url is not None:
                dependabot_url = (
                    repository_url
                    + "/security/dependabot/{}".format(alert["number"])
                )
                description = (
                    "[{}]({})\n".format(dependabot_url, dependabot_url)
                    + description
                )

            finding = Finding(
                title=alert["securityVulnerability"]["advisory"]["summary"],
                test=test,
                description=description,
                severity=self._convert_security(
                    alert["securityVulnerability"].get("severity", "MODERATE")
                ),
                static_finding=True,
                dynamic_finding=False,
                unique_id_from_tool=alert["id"],
            )

            if "vulnerableManifestPath" in alert:
                finding.file_path = alert["vulnerableManifestPath"]

            if "createdAt" in alert:
                finding.date = dateutil.parser.parse(alert["createdAt"])

            if "state" in alert and (
                "FIXED" == alert["state"] or "DISMISSED" == alert["state"]
            ):
                finding.active = False
                finding.is_mitigated = True

            # if the package is present
            if "package" in alert["securityVulnerability"]:
                finding.component_name = alert["securityVulnerability"][
                    "package"
                ].get("name")

            if "references" in alert["securityVulnerability"]["advisory"]:
                finding.references = ""
                for ref in alert["securityVulnerability"]["advisory"][
                    "references"
                ]:
                    finding.references += ref["url"] + "\r\n"

            if "identifiers" in alert["securityVulnerability"]["advisory"]:
                unsaved_vulnerability_ids = list()
                for identifier in alert["securityVulnerability"]["advisory"][
                    "identifiers"
                ]:
                    if identifier.get("value"):
                        unsaved_vulnerability_ids.append(
                            identifier.get("value")
                        )
                if unsaved_vulnerability_ids:
                    finding.unsaved_vulnerability_ids = (
                        unsaved_vulnerability_ids
                    )

            if "cvss" in alert["securityVulnerability"]["advisory"]:
                if (
                    "score"
                    in alert["securityVulnerability"]["advisory"]["cvss"]
                ):
                    score = alert["securityVulnerability"]["advisory"]["cvss"][
                        "score"
                    ]
                    if score is not None:
                        finding.cvssv3_score = score
                if (
                    "vectorString"
                    in alert["securityVulnerability"]["advisory"]["cvss"]
                ):
                    cvss_vector_string = alert["securityVulnerability"][
                        "advisory"
                    ]["cvss"]["vectorString"]
                    if cvss_vector_string is not None:
                        cvss_objects = cvss_parser.parse_cvss_from_text(
                            cvss_vector_string
                        )
                        if len(cvss_objects) > 0:
                            finding.cvssv3 = cvss_objects[0].clean_vector()

            if (
                "cwes" in alert["securityVulnerability"]["advisory"]
                and "nodes"
                in alert["securityVulnerability"]["advisory"]["cwes"]
            ):
                cwe_nodes = alert["securityVulnerability"]["advisory"]["cwes"][
                    "nodes"
                ]
                if cwe_nodes and len(cwe_nodes) > 0:
                    finding.cwe = int(cwe_nodes[0].get("cweId")[4:])

            dupe_key = finding.unique_id_from_tool
            if dupe_key in dupes:
                find = dupes[dupe_key]
                find.nb_occurences += 1
            else:
                dupes[dupe_key] = finding

        return list(dupes.values())

    def _search_vulnerability_alerts(self, data):
        if isinstance(data, list):
            for item in data:
                result = self._search_vulnerability_alerts(item)
                if result:
                    return result
        elif isinstance(data, dict):
            for key in data:
                if key == "vulnerabilityAlerts":
                    return data[key]
                result = self._search_vulnerability_alerts(data[key])
                if result:
                    return result
        return None

    def _convert_security(self, val):
        if val.lower() == "moderate":
            return "Medium"
        else:
            return val.title()
